#pragma once

#include <iterator>
#include <limits>
#include <set>

#if defined(WIN32)
#include <GL/glew.h>
#else
#ifdef NV_USE_OPENGL
#include "GL/gl.h"
#include "GL/glext.h"
#else
#include "nvgles.h"
#endif
#endif

#include "ReadOnlyDatabase.h"
#include "Threading.h"

#if !defined(WIN32)
  #ifdef NV_USE_OPENGL
    extern const char* pNameLibGL;
  #else
    extern const char* pNameLibGLES2;
  #endif
#endif

// Toggle this code block to configure whether errors are checked on every GL statement
#if 0
#define NV_EXEC_GL(_statement)                                                                      \
{                                                                                                   \
    _statement;                                                                                     \
                                                                                                    \
    const GLenum error = glGetError();                                                              \
    static bool ignoreThisError = false;                                                            \
    if (error && !ignoreThisError)                                                                  \
    {                                                                                               \
        NV_ASSERT(false && #_statement);                                                            \
        snprintf(errorBuf, sizeof(errorBuf), "Error 0x%x with statement %s", error, #_statement);   \
        ReportErrorWithMessage(errorBuf, __FILE__, __LINE__);                                       \
    }                                                                                               \
}
#else
#define NV_EXEC_GL(_statement) _statement
#endif

#define NV_CHECK_INIT(_Object)                                                              \
{                                                                                           \
    const GLenum error = glGetError();                                                      \
    if (error)                                                                              \
    {                                                                                       \
        snprintf(errorBuf, sizeof(errorBuf), "Error 0x%x setting up %s", error, #_Object);  \
        ThrowErrorWithMessage(errorBuf, __FILE__, __LINE__);                                \
    }                                                                                       \
}

// Alias the debug usage to the KHR variants when the core debug output functionality is not exposed (OpenGL ES).
// Conditionally alias the debug usage to the KHR variants for Android/QNX/Linux.
#if defined(__ANDROID__) || defined(__QNX__) || defined(__linux__)
#ifndef GL_DEBUG_TYPE_ERROR
#define GL_DEBUG_TYPE_ERROR                  GL_DEBUG_TYPE_ERROR_KHR
#define GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR    GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR_KHR
#define GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR     GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR_KHR
#define GL_DEBUG_TYPE_PORTABILITY            GL_DEBUG_TYPE_PORTABILITY_KHR
#define GL_DEBUG_TYPE_PERFORMANCE            GL_DEBUG_TYPE_PERFORMANCE_KHR
#define GL_DEBUG_TYPE_OTHER                  GL_DEBUG_TYPE_OTHER_KHR
#define GL_DEBUG_TYPE_MARKER                 GL_DEBUG_TYPE_MARKER_KHR
#define GL_DEBUG_TYPE_PUSH_GROUP             GL_DEBUG_TYPE_PUSH_GROUP_KHR
#define GL_DEBUG_TYPE_POP_GROUP              GL_DEBUG_TYPE_POP_GROUP_KHR
#define GL_DEBUG_SOURCE_API                  GL_DEBUG_SOURCE_API_KHR
#define GL_DEBUG_SOURCE_WINDOW_SYSTEM        GL_DEBUG_SOURCE_WINDOW_SYSTEM_KHR
#define GL_DEBUG_SOURCE_SHADER_COMPILER      GL_DEBUG_SOURCE_SHADER_COMPILER_KHR
#define GL_DEBUG_SOURCE_THIRD_PARTY          GL_DEBUG_SOURCE_THIRD_PARTY_KHR
#define GL_DEBUG_SOURCE_APPLICATION          GL_DEBUG_SOURCE_APPLICATION_KHR
#define GL_DEBUG_SOURCE_OTHER                GL_DEBUG_SOURCE_OTHER_KHR
#define GL_DEBUG_SEVERITY_HIGH               GL_DEBUG_SEVERITY_HIGH_KHR
#define GL_DEBUG_SEVERITY_MEDIUM             GL_DEBUG_SEVERITY_MEDIUM_KHR
#define GL_DEBUG_SEVERITY_LOW                GL_DEBUG_SEVERITY_LOW_KHR
#define GL_DEBUG_SEVERITY_NOTIFICATION       GL_DEBUG_SEVERITY_NOTIFICATION_KHR
#define glPushDebugGroup                     glPushDebugGroupKHR
#define glPopDebugGroup                      glPopDebugGroupKHR
#endif
#endif

#ifndef APIENTRY
# ifdef GLAPIENTRY
#  define APIENTRY GLAPIENTRY
# elif defined(GL_APIENTRY)
#  define APIENTRY GL_APIENTRY
# else
#  define APIENTRY
# endif
#endif

extern void APIENTRY DebugCallback(GLenum source, GLenum type, GLuint id, GLenum severity, GLsizei length, const GLchar* message, const GLvoid* userParam);

template <typename T>
T** InitializeListOfLists(int resourceId, std::vector<uint8_t>& retvalData, Serialization::BlobProxy<uint8_t*>& dataProxy);
#define NV_GET_LIST_OF_LISTS(_PtrType, _Handle, _RetValData, _DataProxy) InitializeListOfLists<_PtrType>(_Handle, _RetValData, _DataProxy)

//-----------------------------------------------------------------------------
// GL resource typedefs
//-----------------------------------------------------------------------------
#if defined(_WIN32)
typedef HGLRC Context;
#endif
typedef GLint Uniform;
typedef GLuint UniformBlock;
typedef GLuint ShaderStorageBlock;
typedef GLuint Buffer;
typedef GLuint Shader;
typedef GLuint Program;
typedef GLuint Texture;
typedef GLuint64 TextureHandle;
typedef GLuint VertexArray;
typedef GLuint FrameBuffer;
typedef GLuint ProgramPipeline;
typedef GLuint Query;
typedef GLuint RenderBuffer;
typedef GLuint Sampler;
typedef GLuint TransformFeedback;
typedef unsigned char ClientSideBuffer;
typedef GLuint LegacyProgram;
typedef GLuint Memory;
typedef GLuint Semaphore;
typedef GLuint OVRSwapChainBufferGL;

//-----------------------------------------------------------------------------
// GL indirect structs
//-----------------------------------------------------------------------------
struct glDrawElementsIndirectCommand
{
    GLuint count;
    GLuint drawCount;
    GLuint firstIndex;
    GLuint baseVertex;
    GLuint baseInstance;
};

typedef glDrawElementsIndirectCommand glDrawElementsIndirectCountCommand;
typedef glDrawElementsIndirectCountCommand glDrawElementsIndirectCountARBCommand;

struct glDrawArraysIndirectCommand
{
    GLuint count;
    GLuint instanceCount;
    GLuint first;
    GLuint baseInstance;
};

typedef glDrawArraysIndirectCommand glDrawArraysIndirectCountCommand;
typedef glDrawArraysIndirectCountCommand glDrawArraysIndirectCountARBCommand;

//-----------------------------------------------------------------------------
// Invalid GL object handles
//-----------------------------------------------------------------------------
const Buffer INVALID_BUFFER = (std::numeric_limits<GLuint>::max)();
const FrameBuffer INVALID_FRAMEBUFFER = (std::numeric_limits<GLuint>::max)();
const Program INVALID_PROGRAM = (std::numeric_limits<GLuint>::max)();
const ProgramPipeline INVALID_PROGRAM_PIPELINE = (std::numeric_limits<GLuint>::max)();
const Query INVALID_QUERY = (std::numeric_limits<GLuint>::max)();
const RenderBuffer INVALID_RENDERBUFFER = (std::numeric_limits<GLuint>::max)();
const Sampler INVALID_SAMPLER = (std::numeric_limits<GLuint>::max)();
const Shader INVALID_SHADER = (std::numeric_limits<GLuint>::max)();
const Texture INVALID_TEXTURE = (std::numeric_limits<GLuint>::max)();
const TransformFeedback INVALID_TRANSFORM_FEEDBACK = (std::numeric_limits<GLuint>::max)();
const VertexArray INVALID_VERTEX_ARRAY = (std::numeric_limits<GLuint>::max)();

//-----------------------------------------------------------------------------
// Invalid GL sync objects
//-----------------------------------------------------------------------------
const GLsync INVALID_SYNC = GLsync(0);

//-----------------------------------------------------------------------------
// Invalid GLSL variables
//-----------------------------------------------------------------------------
const ShaderStorageBlock INVALID_SHADER_STORAGE_BLOCK = (std::numeric_limits<GLuint>::max)();
const Uniform INVALID_UNIFORM = (std::numeric_limits<GLint>::max)();
const UniformBlock INVALID_UNIFORM_BLOCK = (std::numeric_limits<GLuint>::max)();

//-----------------------------------------------------------------------------
// GLThreadRunner
//-----------------------------------------------------------------------------
class GLThreadRunner : public ThreadRunner
{
public:
#if defined(_WIN32)
    GLThreadRunner(HANDLE sequenceEvent, HANDLE doneEvent, RunFrameFunc runFrameFunc, ThreadInitFunc initFunc, ThreadReleaseFunc releaseFunc, ResetFrameFunc resetFrameFunc)
        : ThreadRunner(sequenceEvent, doneEvent, runFrameFunc)
        , m_initFunc(initFunc)
        , m_releaseFunc(releaseFunc)
        , m_resetFrameFunc(resetFrameFunc)
    {
    }
#else
    GLThreadRunner(unsigned int threadId, RunFrameFunc runFrameFunc, ThreadInitFunc initFunc, ThreadReleaseFunc releaseFunc, ResetFrameFunc resetFrameFunc)
        : ThreadRunner(threadId, runFrameFunc)
        , m_initFunc(initFunc)
        , m_releaseFunc(releaseFunc)
        , m_resetFrameFunc(resetFrameFunc)
    {
    }
#endif
    virtual void Init() override
    {
        m_initFunc();
    }

    virtual void Release() override
    {
        m_releaseFunc();
    }

private:

    virtual void PostRunThread() override
    {
        // Make sure to let the main thread know we are here and wait for its permission
        // to continue. We also want to make sure that all of the thread contexts are cleared 
        // before running our reset method.
#if defined(_WIN32)
        wglMakeCurrent(NULL, NULL);
#endif
        SignalMainThreadAndWait();
        if (hasFrameReset) {
            m_resetFrameFunc(false);
        }
    }

    ThreadInitFunc m_initFunc;
    ThreadReleaseFunc m_releaseFunc;
    ResetFrameFunc m_resetFrameFunc;
};

//-----------------------------------------------------------------------------
// ConfigSelector
//-----------------------------------------------------------------------------
template <typename T>
class ConfigSelector
{
public:
    inline T GetConfig() const
    {
        return (m_configs.size() > 0 ? *m_configs.begin() : 0);
    }

    inline int GetConfigs(T* configs, int num) const
    {
        int total = std::min(num, int(m_configs.size()));
        int i = 0;
        for (typename std::set<T>::const_iterator it = m_configs.begin(); it != m_configs.end(); ++it)
        {
            configs[i++] = *it;
        }
        return total;
    }

    void IntersectConfigs(const T* configs, int num)
    {
        if (!configs)
        {
            return;
        }

        std::set<T> setConfigs(configs, configs + num);

        // Compute intersections between two sets
        if (m_init)
        {
            m_init = false;
            m_configs.swap(setConfigs);
        }
        else
        {
            std::set<T> intersection;
            std::set_intersection(m_configs.begin(), m_configs.end(), setConfigs.begin(), setConfigs.end(), std::inserter(intersection, intersection.end()));
            m_configs.swap(intersection);
        }
    }

private:
    std::set<T> m_configs;
    bool m_init = true;
};
