//-------------------------------------------------------------------------------
// File: Resources\OpenGLReplay.cpp
//
// Copyright (c) NVIDIA Corporation.  All rights reserved.
//-------------------------------------------------------------------------------

#include "Application.h"
#include "Arguments.h"

#include "OpenGLReplay.h"

#if !defined(WIN32)
#ifdef NV_USE_OPENGL
#include <NvEW_GL.h>
#else
#include <NvEW_GLES2.h>
#endif
#endif

#include <map>
#include <set>

#if !defined(WIN32) && !defined(NV_USE_OPENGL)
    #if !defined(GL_MAP_PERSISTENT_BIT)
        #define GL_MAP_PERSISTENT_BIT GL_MAP_PERSISTENT_BIT_EXT
    #endif
#endif

NoErrorMode noErrorMode = NoErrorMode::ApplicationDefault;

static std::map<GLuint, void*> s_mapPersistentlyMappedBuffers;
static std::set<GLuint> s_setImmutableStorageBuffers;

//------------------------------------------------------------------------------
// OpenGL-specific option parser
//------------------------------------------------------------------------------
static FnParseResults ParseOpenGLArguments(args::ArgumentParser& parser)
{
    static args::Group optOpenGLGroup(parser, "OpenGL-specific options");

    static args::ValueFlag<bool> optNoError(parser, "0 or 1", "Force Enable or Disable KHR_no_error context creation", { "noerror", "no-error" });

    return [=]() {
        if (optNoError) {
            noErrorMode = optNoError.Get() ? NoErrorMode::Enable : NoErrorMode::Disable;
        }

        if (Application::MinimizeCPUOverheadRequested()) {
            // This is where we set any options that we know are safe to enable and minimize the CPU overhead
        }
    };
}

REGISTER_ARGUMENTS(ParseOpenGLArguments);

#if defined(WIN32) || defined(NV_USE_OPENGL)
//------------------------------------------------------------------------------
// GLReplay_MapBuffer
//------------------------------------------------------------------------------
void* GLReplay_MapBuffer(GLuint buffer, GLenum target, GLenum access)
{
    return glMapBuffer(target, access);
}

//------------------------------------------------------------------------------
// GLReplay_MapNamedBuffer
//------------------------------------------------------------------------------
void* GLReplay_MapNamedBuffer(GLuint buffer, GLenum access)
{
    return glMapNamedBuffer(buffer, access);
}

//------------------------------------------------------------------------------
// GLReplay_MapNamedBufferEXT
//------------------------------------------------------------------------------
void* GLReplay_MapNamedBufferEXT(GLuint buffer, GLenum access)
{
    return glMapNamedBufferEXT(buffer, access);
}

//------------------------------------------------------------------------------
// GLReplay_MapNamedBufferRange
//------------------------------------------------------------------------------
void* GLReplay_MapNamedBufferRange(GLuint buffer, GLintptr offset, GLsizeiptr length, GLbitfield access)
{
    const auto itr = s_mapPersistentlyMappedBuffers.find(buffer);
    if (itr != s_mapPersistentlyMappedBuffers.end()) {
        return itr->second;
    }

    auto ptr = glMapNamedBufferRange(buffer, offset, length, access);
    if (ptr && (access & GL_MAP_PERSISTENT_BIT)) {
        s_mapPersistentlyMappedBuffers.insert({ buffer, ptr });
    }

    return ptr;
}

//------------------------------------------------------------------------------
// GLReplay_MapNamedBufferRangeEXT
//------------------------------------------------------------------------------
void* GLReplay_MapNamedBufferRangeEXT(GLuint buffer, GLintptr offset, GLsizeiptr length, GLbitfield access)
{
    const auto itr = s_mapPersistentlyMappedBuffers.find(buffer);
    if (itr != s_mapPersistentlyMappedBuffers.end()) {
        return itr->second;
    }

    auto ptr = glMapNamedBufferRangeEXT(buffer, offset, length, access);
    if (ptr && (access & GL_MAP_PERSISTENT_BIT)) {
        s_mapPersistentlyMappedBuffers.insert({ buffer, ptr });
    }

    return ptr;
}

//------------------------------------------------------------------------------
// GLReplay_UnmapNamedBuffer
//------------------------------------------------------------------------------
void GLReplay_UnmapNamedBuffer(GLuint buffer)
{
    const auto itr = s_mapPersistentlyMappedBuffers.find(buffer);
    if (itr != s_mapPersistentlyMappedBuffers.end()) {
        s_mapPersistentlyMappedBuffers.erase(itr);
    }

    glUnmapNamedBuffer(buffer);
}

//------------------------------------------------------------------------------
// GLReplay_UnmapNamedBufferEXT
//------------------------------------------------------------------------------
void GLReplay_UnmapNamedBufferEXT(GLuint buffer)
{
    const auto itr = s_mapPersistentlyMappedBuffers.find(buffer);
    if (itr != s_mapPersistentlyMappedBuffers.end()) {
        s_mapPersistentlyMappedBuffers.erase(itr);
    }

    glUnmapNamedBufferEXT(buffer);
}

//------------------------------------------------------------------------------
// GLReplay_BufferStorage
//------------------------------------------------------------------------------
void GLReplay_BufferStorage(GLuint buffer, GLenum target, GLsizeiptr size, const GLvoid* data, GLbitfield flags)
{
    const auto itr = s_setImmutableStorageBuffers.find(buffer);
    if (itr != s_setImmutableStorageBuffers.end()) {
        return;
    }

    glBufferStorage(target, size, data, flags);

    s_setImmutableStorageBuffers.insert(buffer);
}

//------------------------------------------------------------------------------
// GLReplay_NamedBufferStorage
//------------------------------------------------------------------------------
void GLReplay_NamedBufferStorage(GLuint buffer, GLsizeiptr size, const GLvoid* data, GLbitfield flags)
{
    const auto itr = s_setImmutableStorageBuffers.find(buffer);
    if (itr != s_setImmutableStorageBuffers.end()) {
        return;
    }

    glNamedBufferStorage(buffer, size, data, flags);

    s_setImmutableStorageBuffers.insert(buffer);
}

//------------------------------------------------------------------------------
// GLReplay_NamedBufferStorageEXT
//------------------------------------------------------------------------------
void GLReplay_NamedBufferStorageEXT(GLuint buffer, GLsizeiptr size, const GLvoid* data, GLbitfield flags)
{
    const auto itr = s_setImmutableStorageBuffers.find(buffer);
    if (itr != s_setImmutableStorageBuffers.end()) {
        return;
    }

    glNamedBufferStorageEXT(buffer, size, data, flags);

    s_setImmutableStorageBuffers.insert(buffer);
}
#endif

//------------------------------------------------------------------------------
// GLReplay_MapBufferOES
//------------------------------------------------------------------------------
void* GLReplay_MapBufferOES(GLuint buffer, GLenum target, GLenum access)
{
    return glMapBufferOES(target, access);
}

//------------------------------------------------------------------------------
// GLReplay_MapBufferRange
//------------------------------------------------------------------------------
void* GLReplay_MapBufferRange(GLuint buffer, GLenum target, GLintptr offset, GLsizeiptr length, GLbitfield access)
{
    const auto itr = s_mapPersistentlyMappedBuffers.find(buffer);
    if (itr != s_mapPersistentlyMappedBuffers.end()) {
        return itr->second;
    }

    auto ptr = glMapBufferRange(target, offset, length, access);
    if (ptr && (access & GL_MAP_PERSISTENT_BIT)) {
        s_mapPersistentlyMappedBuffers.insert({ buffer, ptr });
    }

    return ptr;
}

//------------------------------------------------------------------------------
// GLReplay_UnmapBuffer
//------------------------------------------------------------------------------
void GLReplay_UnmapBuffer(GLuint buffer, GLenum target)
{
    const auto itr = s_mapPersistentlyMappedBuffers.find(buffer);
    if (itr != s_mapPersistentlyMappedBuffers.end()) {
        s_mapPersistentlyMappedBuffers.erase(itr);
    }

    glUnmapBuffer(target);
}

//------------------------------------------------------------------------------
// GLReplay_UnmapBufferOES
//------------------------------------------------------------------------------
void GLReplay_UnmapBufferOES(GLuint buffer, GLenum target)
{
    const auto itr = s_mapPersistentlyMappedBuffers.find(buffer);
    if (itr != s_mapPersistentlyMappedBuffers.end()) {
        s_mapPersistentlyMappedBuffers.erase(itr);
    }

    glUnmapBufferOES(target);
}

//------------------------------------------------------------------------------
// GLReplay_BufferStorageEXT
//------------------------------------------------------------------------------
void GLReplay_BufferStorageEXT(GLuint buffer, GLenum target, GLsizeiptr size, const GLvoid* data, GLbitfield flags)
{
    const auto itr = s_setImmutableStorageBuffers.find(buffer);
    if (itr != s_setImmutableStorageBuffers.end()) {
        return;
    }

    glBufferStorageEXT(target, size, data, flags);

    s_setImmutableStorageBuffers.insert(buffer);
}
